# benchmark_model
# Author:       Everett Harding
# Date Created: 1/26/18
# Description:


import numpy as np
# import os
import datetime
import random
import time
from keras.wrappers.scikit_learn import KerasClassifier
from keras.layers import Dense, Dropout, Average
from keras.models import Model, Sequential
from keras.engine.topology import Input
from sklearn.model_selection import cross_val_score, StratifiedKFold
from sklearn.preprocessing import LabelEncoder, StandardScaler
from sklearn.pipeline import Pipeline
from random import randint

import sys
sys.path.extend([
    '/Users/jack/Documents/Classes/MQP/MQP-ZHAW', '../',
    '/home/tachanka/CS/ZHAW'
])
print(sys.path)

from Utilities.data_utils import load_uci_set, downsample_data, save_metrics_file

model_in_shape = 23
seed = 7
hidden_reduction = .8
dropout_rate = .4

# specific to the autoencoder
enc_hidden_units = randint(1, 46)
enc_epochs = randint(1, 2)
enc_batch = randint(100, 1000)

# used in the basic_FFNN and ensemble training
epochs = randint(1, 46)
batch_size = randint(1, 46)
folds = 5

# for generation of the ensemble networks
activations = ['relu', 'sigmoid']
max_hiddenunits = randint(1, 46)
net_params = {}
num_nets = randint(1, 46)


def showDistribution(binary_vector):
    zeros = 0
    ones = 0
    for n in binary_vector:
        if n > 0:
            ones += 1
        else:
            zeros += 1

    percent_1 = float(ones) / float(len(binary_vector))
    percent_0 = float(zeros) / float(len(binary_vector))
    return percent_0, percent_1


def build_model_ffn():
    model = Sequential()

    model.add(
        Dense(model_in_shape, input_dim=model_in_shape, name='model_input'))
    model.add(Dropout(dropout_rate))
    model.add(Dense(int(model_in_shape * hidden_reduction), activation="relu"))
    model.add(Dense(1, activation="sigmoid"))

    model.compile(
        loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])
    return model


def basic_fnn_benchmark(file):
    global model_in_shape
    global seed

    np.random.seed(seed)
    x_raw, y_raw = load_uci_set()

    # shuffle the dataset,
    # resetting after the first to shuffle y with the same permuation as x
    rnd_state = np.random.get_state()
    np.random.shuffle(x_raw)
    np.random.set_state(rnd_state)
    np.random.shuffle(y_raw)

    e = LabelEncoder()
    e.fit(y_raw)
    y_raw = e.transform(y_raw)

    x, y = downsample_data(x_raw[:-100], y_raw[:-100])

    print("X shape:", x.shape, "\tY Shape: ", y.shape)

    # build the autoencoder
    inputs = Input(shape=(23, ))
    first = Dense(23, name="first")(inputs)
    hidden = Dense(enc_hidden_units, name="hidden")(first)
    outputs = Dense(23, name="outputs")(hidden)

    a_enc = Model(inputs, outputs)
    a_enc.compile(optimizer='adam', loss='mse', metrics=['accuracy'])

    # build the simple network
    estimators = []
    scaler = StandardScaler()
    estimator = KerasClassifier(
        build_fn=build_model_ffn,
        epochs=epochs,
        batch_size=batch_size,
        verbose=0)

    estimators.append(('standardize', scaler))
    estimators.append(('mlp', estimator))
    pipeline = Pipeline(estimators)
    kfold = StratifiedKFold(n_splits=folds, shuffle=True, random_state=seed)

    # =========================================
    # fit the autoencoder and simple network
    start_time = time.clock()
    a_enc.fit(
        x[:-2000],
        x[:-2000],
        epochs=enc_epochs,
        batch_size=enc_batch,
        verbose=0)
    a_enc.evaluate(x[-2000:], x[-2000:], verbose=0)

    j_enc = Model(inputs, hidden)
    j_enc.compile(optimizer='adam', loss='mse', metrics=['accuracy'])

    x = j_enc.predict(x)
    model_in_shape = x.shape[1]

    results = cross_val_score(pipeline, x, y, cv=kfold, verbose=1)
    end_time = time.clock()
    # ===========================================

    # save the Results

    p0, p1 = showDistribution(y)

    print("*****************************", file=file)
    print("*         Parameters        *", file=file)
    print("* ------------------------- *", file=file)
    print("* model_in_shape:\t %4d\t*" % model_in_shape, file=file)
    print("* seed:\t\t\t\t %4d\t*" % seed, file=file)
    print("* hidden_reduction:\t %.2f\t*" % hidden_reduction, file=file)
    print("* dropout_rate:\t\t %.2f\t*" % dropout_rate, file=file)
    print("* epochs:\t\t\t %4d\t*" % epochs, file=file)
    print("* batch_size:\t\t %4d\t*" % batch_size, file=file)
    print("* k_folds:\t\t\t %4d\t*" % folds, file=file)
    print("* enc_hidden:\t\t %4d\t*" % enc_hidden_units, file=file)
    print("* enc_epochs:\t\t %4d\t*" % enc_epochs, file=file)
    print("* enc_batch:\t\t %4d\t*" % enc_batch, file=file)
    print("* \t\t\t\t\t\t\t*", file=file)
    print("* Percent 0:\t\t %2.1f\t*" % (p0 * 100), file=file)
    print("* Percent 1:\t\t %2.1f\t*" % (p1 * 100), file=file)
    print("*****************************", file=file)

    print(
        "Results: %.2f%% (%.2f%%)" % (results.mean() * 100,
                                      results.std() * 100),
        file=file)
    print("Total time: ", end_time - start_time, "seconds", file=file)


def build_ensemble_net(num, m_inputs):
    params = {
        'activation': random.choice(activations),
        'hidden_units1': random.randint(2, max_hiddenunits)
    }

    # add the input layer to the model
    # model = Sequential(name='net%d'%num)
    m = Dense(
        model_in_shape, input_dim=model_in_shape,
        name='model%d_in' % num)(m_inputs)

    # possibly add the first dropout layer
    if random.randint(0, 9) < 5:
        params['dropout1'] = random.random()
        m = Dropout(params['dropout1'])(m)

    # add the first hidden layer
    m = Dense(params['hidden_units1'], activation=params['activation'])(m)

    # possibly add the second dropout layer
    if random.randint(0, 9) < 5:
        params['dropout2'] = random.random()
        m = Dropout(params['dropout2'])(m)

    # possibly add a second hidden layer
    if random.randint(0, 9) < 5:
        params['hidden_units2'] = random.randint(2, int(max_hiddenunits / 2))
        m = Dense(params['hidden_units2'])(m)

    # add a uniform output layer for all ensemble models
    m = Dense(1, activation='sigmoid')(m)

    model = Model(m_inputs, m, name="net#%d" % num)
    # finally compile and return the model
    model.compile(
        optimizer='adam', loss='binary_crossentropy', metrics=['accuracy'])
    return model, params


def build_ensemble():
    global net_params
    global num_nets
    # compile all the nets for the ensemble
    model_input = Input(shape=(23, ))
    nets = []
    for x in range(num_nets):
        net, params = build_ensemble_net(x, model_input)
        nets.append(net)
        net_params[str(x)] = params

    outputs = [net.outputs[0] for net in nets]
    y = Average()(outputs)

    model = Model(model_input, y, name='ensemble')

    model.compile(
        optimizer='adam', loss='binary_crossentropy', metrics=['accuracy'])
    return model


def ensemble_benchmark(file):
    # global model_in_shape
    np.random.seed(5)

    # build the ensemble of networks and cross-validate
    x, y = load_uci_set()

    x, y = downsample_data(x, y)

    # assemble the ensemble classifier
    estimators = []
    scaler = StandardScaler()
    ensemble = KerasClassifier(
        build_fn=build_ensemble,
        epochs=epochs,
        batch_size=batch_size,
        verbose=0)

    save_metrics_file(build_ensemble(), y, x)

    estimators.append(('standardize', scaler))
    estimators.append(('mlp', ensemble))
    pipeline = Pipeline(estimators)

    kfold = StratifiedKFold(n_splits=folds, shuffle=True, random_state=seed)

    # fit and evaluate the ensemble
    start_time = time.clock()
    results = cross_val_score(pipeline, x, y, cv=kfold, verbose=1)

    end_time = time.clock()

    # Save the Results
    # os.chdir('benchmark_results')
    # now = datetime.datetime.now()
    # now_formatted = "ensemble_" + str(now.strftime("%Y-%m-%d-%H:%M:%S"))
    # file = open(now_formatted, mode='a')

    p0, p1 = showDistribution(y)
    print("\n\n\n           Ensemble          ", file=file)
    print("*****************************", file=file)
    print("*         Parameters        *", file=file)
    print("* ------------------------- *", file=file)
    print("* model_in_shape:\t %4d\t*" % model_in_shape, file=file)
    print("* seed:\t\t\t\t %4d\t*" % seed, file=file)
    print("* epochs:\t\t\t %4d\t*" % epochs, file=file)
    print("* batch_size:\t\t %4d\t*" % batch_size, file=file)
    print("* k_folds:\t\t\t %4d\t*" % folds, file=file)
    print("* \t\t\t\t\t\t\t*", file=file)
    print("* Percent 0:\t\t %2.1f\t*" % (p0 * 100), file=file)
    print("* Percent 1:\t\t %2.1f\t*" % (p1 * 100), file=file)
    print("*****************************", file=file)

    print(
        "Results: %.2f%% (%.2f%%)" % (results.mean() * 100,
                                      results.std() * 100),
        file=file)
    print("Total time: ", end_time - start_time, "seconds", file=file)

    for num in net_params:
        print("%s:\t" % num, net_params[num], file=file)


if __name__ == '__main__':
    # Save the Results
    # os.chdir('benchmark_results')
    now = datetime.datetime.now()
    now_formatted = "both_" + str(now.strftime("%Y-%m-%d-%H:%M:%S"))
    file = open(now_formatted, mode='a')

    basic_fnn_benchmark(file)
    ensemble_benchmark(file)
