#!/bin/env python3
# SBATCH -p short
# SBATCH -C V100
# SBATCH --gres=gpu:2
# SBATCH -o %x-%j.out
# SBATCH -J Randomized-Parameter-Test
# SBATCH --error=whoops.err
# optimizer.py
# Author:       Luke Buquicchio, Everett Harding

# Date Created: 1/30/18
# Description: An optimizer for our ensemble.
# TODO fix the number of predictors each ensemble sees to sqrt(cols)
# TODO is there some particularly

use_Trees = True

if use_Trees:
    print("Randomized-Parameter-Test-Trees")

else:
    print("Randomized-Parameter-Test")

import random
import time
import sys

import numpy as np
#import matplotlib.pyplot as plt
from keras.models import Sequential
from keras.layers import Dense, Dropout
from keras import metrics
import keras.utils as k_utils
from keras.utils import np_utils
from keras.backend import clear_session
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import confusion_matrix
from sklearn.decomposition import PCA

sys.path.extend(['/home/ejharding/work/MQP-ZHAW',
                 '/home/ljbuquiccio/MQP/MQP-ZHAW',
                 '/Users/jack/Documents/Class/MQP/MQP-ZHAW',
                 '../'])

from Utilities import ensemble_utils
from Utilities import data_utils
from adapter import Adapter
from sklearn.ensemble import RandomForestClassifier


np.set_printoptions(threshold=np.nan)

try:
    data = np.loadtxt('../Data/UCI_Credit_Card.csv', delimiter=',', skiprows=1)
except:
    data = np.loadtxt('Data/UCI_Credit_Card.csv', delimiter=',', skiprows=1)

cat_columns = [False, False, True, True, True, False, True, True, True, True, True, True, False, False, False, False, False, False, False, False, False, False, False, False, True]
cat_ranges =  [0,     0,     2,    4,    3,    0,     10,   10,   10,   10,   10,   10,   0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     2]
reg_columns = np.logical_not(cat_columns)

start_time = time.clock()

x = data[:, 1:-1]
reg_columns = reg_columns[1:-1]
cat_columns = cat_columns[1:-1]
cat_ranges = cat_ranges[1:-1]
y = data[:, -1]
x = data_utils.normalize(x, reg_columns)
x, y = data_utils.downsample_data(x, y)
x_train_test, x_vault, y_train_test, y_vault = data_utils.split_data(x, y, .7)
X, Y = x, y

zeros = 0
ones = 0
avg_ensemble_loss = 0.0
use_PCA = False
use_Trees = False
cvscores = []
losses = []
ratios = []
histories = []


# Ensemble Variables
ensemble_size = random.randint(1, 75)
ensemble_batch_size = random.randint(1000, 2000)
ensemble_epochs = random.randint(1, 2)
net_params = {'batch_size': ensemble_batch_size, 'epochs': ensemble_epochs}
ensemble_keep_feature_threshold = random.random()
while ensemble_keep_feature_threshold == 0.:
    ensemble_keep_feature_threshold = random.random()


# Adapter Variables
adapter_out_shape = 2
adapter_activations = ['softmax']
adapter_losses = ['categorical_crossentropy']
activation = random.choice(adapter_activations)
loss = random.choice(adapter_losses)

adapter_hidden_layer_1_size = random.randrange(int(ensemble_size * 10 - ensemble_size * 5),
                                               int(ensemble_size * 10 + ensemble_size * 5))
adapter_hidden_layer_2_size = random.randrange(1,
                                               int(adapter_hidden_layer_1_size / 10))
loss_function = random.choice(adapter_losses)
adapter_layers = [adapter_hidden_layer_1_size]
if not adapter_hidden_layer_2_size == 0:
    adapter_layers.append(adapter_hidden_layer_2_size)
adapter_layers_dict = ensemble_utils.build_hidden_layer_dict(adapter_layers)


# PCA
pca = PCA(n_components="mle", svd_solver="full")


# K Fold Cross Validation
k_fold = StratifiedKFold(n_splits=10, shuffle=True, random_state=42)
fold_num = 1

# Set the random state to force the ensemble to be the same in every fold of the CV
rand_state_start = np.random.get_state()
for train, test in k_fold.split(x_train_test, y_train_test):
    ensemble_evaluation_string = "Start Fold: %d _-_-_-_-_-_-_-_-_-\n" % fold_num
    np.random.set_state(rand_state_start)

    # Variables for training and testing data
    x_train, x_test = x_train_test[train], x_train_test[test]

    # Keeps track of the ratio of labels in the testing set
    for label in y_train_test[test]:
        if label == 0:
            zeros += 1
        else:
            ones += 1
    ratio = ones/zeros

    # Convert the labels to categorical vectors for Keras
    y_train = k_utils.to_categorical(y_train_test[train].astype(int))
    y_test = k_utils.to_categorical(y_train_test[test].astype(int))

    # # Just a tree
    # forest = RandomForestClassifier(n_estimators=100, max_features=None, n_jobs=-1)
    # forest.fit(x_train, y_train)
    # score = forest.score(x_test, y_test)
    # print(score)

    # Ensemble building and training
    ensemble = ensemble_utils.build_ensemble(ensemble_size,
                                             x_train,
                                             net_params,
                                             cat_columns,
                                             ensemble_keep_feature_threshold)

    # Use the ensemble to transform
    x_train_expanded = ensemble_utils.encode(x_train, ensemble)
    # Expand the testing data be abstracting the penultimate layer of the ensemble members
    x_test_expanded = ensemble_utils.encode(x_test, ensemble)

    if use_PCA:
        # TODO Check that this is being done right before implementation
        pca.fit(x_train_expanded)
        x_train_expanded = pca.transform(x_train_expanded)
        x_test_expanded = pca.transform(x_train_expanded)

    if use_Trees:
        adapter = RandomForestClassifier(n_estimators=100, max_features=None, n_jobs=-1)
        adapter.fit(x_train_expanded, y_train)
        score = adapter.score(x_test_expanded, y_test)
        cvscores.append(score * 100)
        print(score)

    else:
        # The number of columns in the expanded training set
        adapter_in_shape = x_train_expanded.shape[1]
        layer2_shape = adapter_in_shape / 2
        layer3_shape = layer2_shape / 2

        # To insure there isn't a bug in the adapter class, lets build one from scratch
        adapter = Sequential()
        adapter.add(Dense(adapter_hidden_layer_1_size,
                          input_dim=adapter_in_shape,
                          bias_initializer='zeros',
                          kernel_initializer='random_uniform',
                          activation='relu',
                          name="adapter_input_layer"))
        adapter.add(Dropout(.3))
        adapter.add(Dense(adapter_hidden_layer_2_size,
                          input_dim=adapter_hidden_layer_1_size,
                          bias_initializer='zeros',
                          kernel_initializer='random_uniform',
                          activation='relu',
                          name="adapter_hidden_layer"))
        adapter.add(Dropout(.3))
        adapter.add(Dense(adapter_out_shape,
                          input_dim=adapter_hidden_layer_2_size,
                          bias_initializer='zeros',
                          kernel_initializer='random_uniform',
                          activation='softmax',
                          name="adapter_output_layer"))

        adapter.compile(loss=loss, optimizer='adam', metrics=['accuracy', metrics.mae, metrics.cosine])

        # # Adapter building, training, and evaluation
        # adapter = Adapter(adapter_in_shape,
        #                   hidden_units=adapter_layers_dict,
        #                   activation=activation,
        #                   loss=loss_function,
        #                   num_gpus=0,
        #                   verbose=0,
        #                   out_shape=adapter_out_shape)

        # train the adapter
        history = adapter.fit(x_train_expanded, y_train, epochs=150, verbose=0)

        # print the history (to get the loss and accuracy through the epochs"
        ensemble_evaluation_string += 'Adapter Training History:\n'
        for elem in history.history:
            ensemble_evaluation_string += elem + ': ' + str(history.history[elem]) + '\n'

        score = adapter.evaluate(x_test_expanded, y_test)
        data_utils.additional_metrics(adapter, y_test, x_test_expanded)

        ensemble_evaluation_string += "\nAdapter Evaluate Results:\n"
        for x in range(len(score)):
            ensemble_evaluation_string += '%s: %f\n' % (adapter.model.metrics_names[x], score[x])

        cvscores.append(score[1] * 100)
        losses.append(score[0])
        histories.append(history)

    ratios.append(ratio)

    # prep lists for confusion matrix
    preds = [float(1. - round(pred[0])) for pred in adapter.predict(x_test_expanded)]
    ys = [float(1. - ex[0]) for ex in y_test]
    con_mat = confusion_matrix(ys, preds)

    # ensemble evaluation output
    ensemble_evaluation_string += "\nEnsemble Nets:\n"
    ensemble_results = {}
    for entry in ensemble.values():
        avg_loss = 0
        for net, params, mask, label in entry:
            # unpack the tuple and use it to get some info
            is_cat = cat_columns[label]
            if is_cat:
                lb = params['lb']
                # lb.fit(test_f[:,label])
                labels = lb.transform(x_test[:, label])
                if lb.y_type_ == 'binary':
                    labels = np_utils.to_categorical(labels)
            else:
                labels = x_test[:, label]

            result = net.evaluate(x_test[:, mask], labels, verbose=0)

            if is_cat:
                ensemble_evaluation_string += "Target %4d:\tAccy: %.4f\tStd Dev: %.4f\n" % (label, result[1], x_test[:, label].std())
                avg_loss += result[0]
            else:
                ensemble_evaluation_string += "Target %4d:\tLoss: %.4f\tStd Dev: %.4f\n" % (label, result[0], x_test[:, label].std())
                avg_loss += result[0]

            if label in ensemble_results.keys():
                ensemble_results[label].append((result, params, mask))
            else:
                ensemble_results[label] = [(result, params, mask)]
        ensemble_evaluation_string += "Avg: %.4f\n\n" % (avg_loss / len(entry))
    ensemble_evaluation_string += 'Data Ratio: %f\n' % ratio

    # print the confusion matrix and the classification metrics
    ensemble_evaluation_string += 'Adapter Confusion Matrix:\n| TP | FN |\n| FP | TN |\n' + str(con_mat) + '\n\n'
    precision = float(con_mat[0, 0] / (con_mat[:, 0].sum()))
    recall = float(con_mat[0, 0] / con_mat[0, :].sum())
    f1 = 2 * ((precision * recall) / (precision + recall))
    ensemble_evaluation_string += "Precision: %ff\nRecall: %f\nF1 Score: %f\n" % (precision, recall, f1)
    ensemble_evaluation_string += "%d =============================\n\n" % fold_num
    # Print the data
    print(ensemble_evaluation_string, flush=True)
    fold_num += 1

    # clear the keras backend so we can start a-fresh
    clear_session()


end_time = time.clock()

if use_Trees:
    print(
        "Ensemble Size: %d\nAvg Accuracy: %s\nAvg Data Ratio: %s\nTotal Time: "
        % (ensemble_size,
           str(np.mean(cvscores)),
           str(np.mean(ratios))),
        end_time - start_time,
        flush=True)

else:
    print("Ensemble Size: %d\nActivation: %s\nAdapter Hidden 1 Size: %d\nAdapter Hidden 2 Size: %d\nLoss Function: %s\nAvg Loss: %s\nAvg Accuracy: %s\nAvg Data Ratio: %s\nTotal Time: "
          % (ensemble_size,
             activation,
             adapter_hidden_layer_1_size,
             adapter_hidden_layer_2_size,
             loss_function,
             str(np.mean(losses)),
             str(np.mean(cvscores)),
             str(np.mean(ratios))),
          end_time - start_time,
          flush=True)
