import numpy as np
import time
from keras.models import Sequential
from keras.layers import LSTM, Dropout, Dense, Activation


# converts result into nparray, shuffles train data, and the separates them
# method opens file, separates into sections for data (train) and target (test)
# into x and y.
def load_data(filename, seq_len, normalise_window):
    f = open(filename, 'rb').read()
    data = f.decode().split('\n')

    sequence_length = seq_len + 1
    result = []

    # separate parts of the dataset as the y or target.
    for index in range(len(data) - sequence_length):
        result.append(data[index:index + sequence_length])

    if normalise_window:
        result = normalise_windows(result)

    result = np.array(result)  # convert to np array.

    row = round(0.9 * result.shape[0])
    train = result[:int(row), :]
    np.random.shuffle(train)
    x_train = train[:, :-1]
    y_train = train[:, -1]
    y_test = result[int(row):, :-1]
    x_test = result[int(row):, :-1]

    x_train = np.reshape(x_train, (x_train.shape[0], x_train.shape[1], 1))
    x_test = np.reshape(x_test, (x_test.shape[0], x_test.shape[1], 1))

    return [x_train, y_train, x_test, y_test]


# this method normalizes each timeframe (datapoint) before we run it through
# our model.
def normalise_windows(result):
    normalised_data = []
    for window in result:
        normalised_window = [((float(p) / float(window[0])) - 1)
                             for p in window]
        normalised_data.append(normalised_window)

    return normalised_data


# this method builds the neural network and compiles the model based
# on the size specifications.
def build_model(layers):
    model = Sequential()

    model.add(
        LSTM(
            input_shape=layers[0], output_dim=layers[1],
            return_sequences=True))
    model.add(Dropout(0.2))

    model.add(LSTM(layers[2], return_sequences=False))
    model.add(Dropout(0.2))
    model.add(Dense(output_dim=layers[3]))
    model.add(Activation("linear"))

    start = time.time()
    model.compile(loss="mse", optimizer="rmsprop")
    print("> Compilation Time: ", time.time() - start)
    return model
