#!/bin/env python3
# SBATCH -p short
# SBATCH -C V100
# SBATCH --gres=gpu:2
# SBATCH -o %x-%j.out
# SBATCH -J Correlation-Test

# optimizer.py
# Author:       Luke Buquicchio

# Date Created: 1/30/18
# Description: An optimizer for our ensemble.
print("Correlation-Test")


import random
import time
import sys

import numpy as np

sys.path.append('../')

from Utilities import ensemble_utils
from Utilities import data_utils
from Legacy.model_tester import optimizable_main


data = np.loadtxt('../Data/UCI_Credit_Card.csv', delimiter=',', skiprows=1)


start_time = time.clock()

x = data[:, 0:-2]
y = data[:, -1]
x = data_utils.normalize(x)
x_train, x_test, y_train, y_test = data_utils.split_data(x, y, .2)

adapter_activations = ['softmax', 'sigmoid', 'tanh']
adapter_losses = ['categorical_crossentropy']
# 'categorical_hinge',,'binary_crossentropy'

# ensemble_size = random.randrange(10, 100)
ensemble_size = 83
activation = random.choice(adapter_activations)
hidden_layer_1_size = random.randrange(int(ensemble_size * 10 - ensemble_size * 5),
                                       int(ensemble_size * 10 + ensemble_size * 5))
hidden_layer_2_size = random.randrange(int(hidden_layer_1_size / 10), int(hidden_layer_1_size / 2))
hidden_layer_3_size = random.randrange(int(hidden_layer_2_size / 10), int(hidden_layer_2_size / 2))
loss_function = random.choice(adapter_losses)

ensemble = ensemble_utils.generate_ff_ensemble(x_train, num_nets=ensemble_size)
x0 = np.array([data, ensemble, 'sigmoid', 573, 129, 14, 'categorical_crossentropy'])
metrics, ratio, average_ensemble_testing_accuracy, average_ensemble_testing_loss = optimizable_main(x0)

loss = metrics[0]
accuracy = metrics[1]

end_time = time.clock()

print(83, ", ",
      'sigmoid', ", ",
      573, ", ",
      129, ", ",
      14, ", ",
      "categorical_crossentropy", ", ",
      loss, ", ",
      accuracy, ", ",
      ratio, ", ",
      average_ensemble_testing_accuracy, ", ",
      end_time - start_time, ", ",
      average_ensemble_testing_loss,
      flush=True)

# save the Results
with open('./Results/results.csv', 'a', newline='\n') as csvfile:
    print(83, ", ",
          'sigmoid', ", ",
          573, ", ",
          129, ", ",
          14, ", ",
          "categorical_crossentropy", ", ",
          loss, ", ",
          accuracy, ", ",
          ratio, ", ",
          average_ensemble_testing_accuracy, ", ",
          end_time - start_time, ", ",
          average_ensemble_testing_loss,
          file=csvfile)
