#!/bin/env python3
# SBATCH -p short
# SBATCH -C V100
# SBATCH --gres=gpu:2
# SBATCH -o %x-%j.out
# SBATCH -J Randomized-Parameter-Test
# SBATCH --error=whoops.err
# optimizer.py
# Author:       Luke Buquicchio, Everett Harding

# Date Created: 1/30/18
# Description: An optimizer for our ensemble.
print("Randomized-Parameter-Test")
# TODO implement proper path stuff

import random
import time
import sys

import numpy as np
#import matplotlib.pyplot as plt
import keras.utils as k_utils
from keras.utils import np_utils
from keras.backend import clear_session
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import confusion_matrix
from sklearn.decomposition import PCA

sys.path.extend(['/home/ejharding/work/MQP-ZHAW', '../','/home/ljbuquiccio/MQP/MQP-ZHAW', '/Users/jack/Documents/Class/MQP/MQP-ZHAW'])

from Utilities import ensemble_utils
from Utilities import data_utils
from adapter import Adapter


np.set_printoptions(threshold=np.nan)

try:
    data = np.loadtxt('../Data/UCI_Credit_Card.csv', delimiter=',', skiprows=1)
except:
    data = np.loadtxt('Data/UCI_Credit_Card.csv', delimiter=',', skiprows=1)
cat_columns = [False, False, True, True, True, False, True, True, True, True, True, True, False, False, False, False, False, False, False, False, False, False, False, False, True]
reg_columns = np.logical_not(cat_columns)
cat_ranges =  [0,     0,     2,    4,    3,    0,     10,   10,   10,   10,   10,   10,   0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     2]

start_time = time.clock()

x = data[:, 1:-1]
reg_columns = reg_columns[1:-1]
cat_columns = cat_columns[1:-1]
cat_ranges = cat_ranges[1:-1]
y = data[:, -1]
x = data_utils.normalize(x, reg_columns)
x, y = data_utils.downsample_data(x, y)
X, X_vault, Y, Y_vault = data_utils.split_data(x, y, .7)
X, Y = x, y

zeros = 0
ones = 0
avg_ensemble_loss = 0.0
avg_ensemble_accuracy = 0.0
use_PCA = False
cvscores = []
losses = []
accuracies = []
ratios = []
histories = []


# Ensemble Variables
ensemble_size = random.randint(10, 200)
ensemble_batch_size = 2000
ensemble_epochs = 2
net_params = {'batch_size': ensemble_batch_size, 'epochs': ensemble_epochs}
ensemble_keep_feature_threshold = 1
while ensemble_keep_feature_threshold == 0.:
    ensemble_keep_feature_threshold = random.random()


# Adapter Variables
adapter_out_shape = 2
adapter_activations = ['softmax']
adapter_losses = ['categorical_crossentropy']
activation = random.choice(adapter_activations)
adapter_hidden_layer_1_size = 200
adapter_hidden_layer_2_size = 75
adapter_hidden_layer_3_size = 25
#adapter_hidden_layer_1_size = random.randrange(int(ensemble_size * 10 - ensemble_size * 5),
#                                               int(ensemble_size * 10 + ensemble_size * 5))
#adapter_hidden_layer_2_size = random.randrange(int(adapter_hidden_layer_1_size / 10),
#                                               int(adapter_hidden_layer_1_size / 2))
# adapter_hidden_layer_3_size = random.randrange(int(adapter_hidden_layer_2_size / 10),
#                                                int(adapter_hidden_layer_2_size / 2))
loss_function = random.choice(adapter_losses)
adapter_layers = [adapter_hidden_layer_1_size]
if not adapter_hidden_layer_2_size == 0:
    adapter_layers.append(adapter_hidden_layer_2_size)
if not adapter_hidden_layer_3_size == 0:
    adapter_layers.append(adapter_hidden_layer_3_size)
adapter_layers_dict = ensemble_utils.build_hidden_layer_dict(adapter_layers)


# PCA
pca = PCA(n_components="mle", svd_solver="full")


# K Fold Cross Validation
k_fold = StratifiedKFold(n_splits=10, shuffle=True, random_state=42)
fold_num = 1

# Set the random state to force the ensemble to be the same in every fold of the CV
rand_state_start = np.random.get_state()
for train, test in k_fold.split(X, Y):
    ensemble_evaluation_string = "Start Fold: %d _-_-_-_-_-_-_-_-_-\n" % fold_num
    np.random.set_state(rand_state_start)

    # Variables for training and testing data
    x_train, x_test = X[train], X[test]

    # Keeps track of the ratio of labels in the testing set
    for label in Y[test]:
        if label == 0:
            zeros += 1
        else:
            ones += 1
    ratio = ones/zeros

    # Convert the labels to categorical vectors for Keras
    y_train = k_utils.to_categorical(Y[train].astype(int))
    y_test = k_utils.to_categorical(Y[test].astype(int))

    # Ensemble building and training
    ensemble = ensemble_utils.build_ensemble(ensemble_size,
                                             x_train,
                                             net_params,
                                             cat_columns,
                                             ensemble_keep_feature_threshold)

    # Use the ensemble to transform
    x_train_expanded = ensemble_utils.encode(x_train, ensemble)
    if use_PCA:
        pca.fit(x_train_expanded)
        x_train_expanded = pca.transform(x_train_expanded)

    # The number of columns in the expanded training set
    adapter_in_shape = x_train_expanded.shape[1]

    # Adapter building, training, and evaluation
    adapter = Adapter(adapter_in_shape,
                      hidden_units=adapter_layers_dict,
                      activation=activation,
                      loss=loss_function,
                      num_gpus=0,
                      verbose=0,
                      out_shape=adapter_out_shape)

    # train the adapter
    history = adapter.fit(x_train_expanded, y_train, epochs=150)

    # print the history (to get the loss and accuracy through the epochs"
    ensemble_evaluation_string += 'Adapter Training History:\n'
    for elem in history.history:
        ensemble_evaluation_string += elem + ': ' + str(history.history[elem]) + '\n'

    # Expand the testing data be abstracting the penultimate layer of the ensemble members
    x_test_expanded = ensemble_utils.encode(x_test, ensemble)
    if use_PCA:
        pca.fit(x_test_expanded)
        x_train_expanded = pca.transform(x_test_expanded)

    score = adapter.evaluate(x_test_expanded, y_test)

    ensemble_evaluation_string += "\nAdapter Evaluate Results:\n"
    for x in range(len(score)):
        ensemble_evaluation_string += '%s: %f\n' % (adapter.model.metrics_names[x], score[x])


    # Save some data
    cvscores.append(score[1] * 100)
    losses.append(score[0])
    accuracies.append(score[1])
    ratios.append(ratio)
    histories.append(history)

    # prep lists for confusion matrix
    preds = [float(1. - round(pred[0])) for pred in adapter.predict(x_test_expanded)]
    ys = [float(1. - ex[0]) for ex in y_test]
    con_mat = confusion_matrix(ys, preds)

    # ensemble evaluation output
    ensemble_evaluation_string += "\nEnsemble Nets:\n"
    ensemble_results = {}
    for entry in ensemble.values():
        avg_loss = 0
        for net, params, mask, label in entry:
            # unpack the tuple and use it to get some info
            is_cat = cat_columns[label]
            if is_cat:
                lb = params['lb']
                # lb.fit(test_f[:,label])
                labels = lb.transform(x_test[:, label])
                if lb.y_type_ == 'binary':
                    labels = np_utils.to_categorical(labels)
            else:
                labels = x_test[:, label]

            result = net.evaluate(x_test[:, mask], labels, verbose=0)

            if is_cat:
                ensemble_evaluation_string += "Target %4d:\tAccy: %.4f\tStd Dev: %.4f\n" % (label, result[1], x_test[:, label].std())
                avg_loss += result[0]
            else:
                ensemble_evaluation_string += "Target %4d:\tLoss: %.4f\tStd Dev: %.4f\n" % (label, result[0], x_test[:, label].std())
                avg_loss += result[0]

            if label in ensemble_results.keys():
                ensemble_results[label].append((result, params, mask))
            else:
                ensemble_results[label] = [(result, params, mask)]
        ensemble_evaluation_string += "Avg: %.4f\n\n" % (avg_loss / len(entry))
    ensemble_evaluation_string += 'Data Ratio: %f\n' % ratio

    # print the confusion matrix and the classification metrics
    ensemble_evaluation_string += 'Adapter Confusion Matrix:\n| TP | FN |\n| FP | TN |\n' + str(con_mat) + '\n\n'
    precision = float(con_mat[0, 0] / (con_mat[:, 0].sum()))
    recall = float(con_mat[0, 0] / con_mat[0, :].sum())
    f1 = 2 * ((precision * recall) / (precision + recall))
    ensemble_evaluation_string += "Precision: %ff\nRecall: %f\nF1 Score: %f\n" % (precision, recall, f1)
    ensemble_evaluation_string += "%d =============================\n\n" % fold_num
    # Print the data
    print(ensemble_evaluation_string, flush=True)
    fold_num += 1

    # clear the keras backend so we can start a-fresh
    clear_session()


end_time = time.clock()
print("Ensemble Size: %d\nActivation: %s\nAdapter Hidden 1 Size: %d\nAdapter Hidden 2 Size: %d\nAdapter Hidden 3 Size: %d\nLoss Function: %s\nAvg Loss: %s\nAvg Accuracy: %s\nAvg Data Ratio: %s\nTotal Time: "
      % (ensemble_size,
         activation,
         adapter_hidden_layer_1_size,
         adapter_hidden_layer_2_size,
         adapter_hidden_layer_3_size,
         loss_function,
         str(np.mean(losses)),
         str(np.mean(accuracies)),
         str(np.mean(ratios))),
      end_time - start_time,
      flush=True)
