# optimizer.py
# Author:       Luke Buquicchio

# Date Created: 2/19/18
# Description: Read in the results from the testing and graph

import glob

from matplotlib import pyplot as plt
import numpy as np


# This function will tell if a character is an int
def represents_int(i):
    try:
        int(i)
        return True
    except ValueError:
        return False


# Get a list of files to read data from
list_of_files = glob.glob('../Results/*.out')

randomized_parameter_data = np.zeros((1, 12))
correlation_data = np.zeros((1, 12))

for file in list_of_files:

    with open(file, newline='') as outfile:
        lines = outfile.readlines()
        if len(lines) > 1:
            # Sometimes the label appears on the first line, sometimes the second
            result_type = lines[0].strip()
            result_type2 = lines[1].strip()
        else:
            # In the case of a blank file
            continue

        if len(lines) > 1:
            for line in lines:
                # Remove unnecessary characters
                line = line.replace(" ", "")
                line = line.replace("\n", "")

                # The current output files have garbage lines from warnings. The lines containing data uniquely have
                #   ints at the first and last index of the line.
                if represents_int(line[0]) and represents_int(line[-1]):
                    data = line.split(",")

                    # Since we added the printout of the average ensemble loss after the fact, some data has improper
                    #   dimensions. This rectifies that.
                    if len(data) < 12:
                        data.append(0)

                    # Add the new data to the matrix
                    if result_type == "Randomized-Parameter-Test" or result_type2 == "Randomized-Parameter-Test":
                        randomized_parameter_data = np.vstack([randomized_parameter_data, data])

                    elif result_type == "Correlation-Test" or result_type2 == "Correlation-Test":
                        correlation_data = np.vstack([correlation_data, data])

# The first row is zero from declaring the numpy array
randomized_parameter_data = x = np.delete(randomized_parameter_data, 0, axis=0)
correlation_data = np.delete(correlation_data, 0, axis=0)

print("Randomized Parameter Data size: ", randomized_parameter_data.shape[0])
print("Correlation Data size:          ", correlation_data.shape[0])


# Graphing the results for the randomized performance
ensemble_size_rpd =       randomized_parameter_data[:, 0].astype(int)
adapter_activation_rpd =  randomized_parameter_data[:, 1].astype(str)
hidden_layer_1_size_rpd = randomized_parameter_data[:, 2].astype(int)
hidden_layer_2_size_rpd = randomized_parameter_data[:, 3].astype(int)
hidden_layer_3_size_rpd = randomized_parameter_data[:, 4].astype(int)
loss_function_rpd =       randomized_parameter_data[:, 5].astype(str)
adapter_loss_rpd =        randomized_parameter_data[:, 6].astype(float)
adapter_accuracy_rpd =    randomized_parameter_data[:, 7].astype(float)
category_ratio_rpd =      randomized_parameter_data[:, 8].astype(float)
ensemble_accuracy_rpd =   randomized_parameter_data[:, 9].astype(float)
total_running_time_rpd =  randomized_parameter_data[:, 10].astype(float)
ensemble_loss_rpd =       randomized_parameter_data[:, 11].astype(float)

# The first figure
fig1_rpd, ((ax11_rpd, ax12_rpd), (ax21_rpd, ax22_rpd)) = plt.subplots(2, 2)

ax11_rpd.scatter(ensemble_accuracy_rpd, adapter_accuracy_rpd, s=5)
ax11_rpd.set_xlabel("Ensemble Accuracy")
ax11_rpd.set_ylabel("Model Accuracy")
ax11_rpd.set_title("Ensemble Accuracy vs Model Accuracy")

ax12_rpd.scatter(category_ratio_rpd, adapter_accuracy_rpd, s=5)
ax12_rpd.set_xlabel("Category Ratio")
ax12_rpd.set_ylabel("Model Accuracy")
ax12_rpd.set_title("Category Ratio vs Model Accuracy")

ax21_rpd.scatter(adapter_loss_rpd, adapter_accuracy_rpd, s=5)
ax21_rpd.set_xlabel("Adapter Loss")
ax21_rpd.set_ylabel("Model Accuracy")
ax21_rpd.set_title("Model Loss vs Model Accuracy")

ax22_rpd.scatter(total_running_time_rpd, adapter_accuracy_rpd, s=5)
ax22_rpd.set_xlabel("Total Run Time")
ax22_rpd.set_ylabel("Model Accuracy")
ax22_rpd.set_title("Total Run Time vs Model Accuracy")

plt.tight_layout(pad=1.5, w_pad=1.5, h_pad=1.5)
fig1_rpd.set_dpi(500)


# The second figure
fig2_rpd, (ax1_2rpd, ax2_2rpd) = plt.subplots(2, 1)

rows_rpd = randomized_parameter_data.shape[0]
clust1_mask_rpd = [False] * rows_rpd
clust2_mask_rpd = [False] * rows_rpd
clust3_mask_rpd = [False] * rows_rpd

# Separate the data based on model performance.
# From here on, clust3 represents all of the models that actually gave real results.
for i in range(rows_rpd):
    if adapter_accuracy_rpd[i] < 0.5:
        clust1_mask_rpd[i] = True
    elif 0.5 <= adapter_accuracy_rpd[i] < 0.6:
        clust2_mask_rpd[i] = True
    elif adapter_accuracy_rpd[i] >= 0.6:
        clust3_mask_rpd[i] = True

clust1_adapter_accuracy_rpd = adapter_accuracy_rpd[clust1_mask_rpd]
clust2_adapter_accuracy_rpd = adapter_accuracy_rpd[clust2_mask_rpd]
clust3_adapter_accuracy_rpd = adapter_accuracy_rpd[clust3_mask_rpd]

clust1_ensemble_accuracy_rpd = ensemble_accuracy_rpd[clust1_mask_rpd]
clust2_ensemble_accuracy_rpd = ensemble_accuracy_rpd[clust2_mask_rpd]
clust3_ensemble_accuracy_rpd = ensemble_accuracy_rpd[clust3_mask_rpd]

clust1_hidden_layer_1_rpd = hidden_layer_1_size_rpd[clust1_mask_rpd]
clust2_hidden_layer_1_rpd = hidden_layer_1_size_rpd[clust2_mask_rpd]
clust3_hidden_layer_1_rpd = hidden_layer_1_size_rpd[clust3_mask_rpd]

# ax11_2rpd.scatter(clust1_ensemble_accuracy_rpd, clust1_adapter_accuracy_rpd, s=5, c='blue')
# ax11_2rpd.scatter(clust2_ensemble_accuracy_rpd, clust2_adapter_accuracy_rpd, s=5, c='red')
ax1_2rpd.scatter(clust3_ensemble_accuracy_rpd, clust3_adapter_accuracy_rpd, s=5, c='green')
ax1_2rpd.set_xlabel("Ensemble Accuracy")
ax1_2rpd.set_ylabel("Model Accuracy")
ax1_2rpd.set_title("Ensemble Accuracy vs Model Accuracy")

# ax12_2rpd.scatter(clust1_hidden_layer_1_rpd, clust1_adapter_accuracy_rpd, s=5, c='blue')
# ax12_2rpd.scatter(clust2_hidden_layer_1_rpd, clust2_adapter_accuracy_rpd, s=5, c='red')
ax2_2rpd.scatter(clust3_hidden_layer_1_rpd, clust3_adapter_accuracy_rpd, s=5, c='green')
ax2_2rpd.set_xlabel("Hidden Layer 1 Size")
ax2_2rpd.set_ylabel("Model Accuracy")
ax2_2rpd.set_title("Hidden Layer 1 Size vs Model Accuracy")

plt.tight_layout(pad=1.5, w_pad=1.5, h_pad=1.5)
fig2_rpd.set_dpi(500)


# The third figure
fig3_rpd, (ax1_3rpd, ax2_3rpd) = plt.subplots(2, 1)

clust1_hidden_layer_2_rpd = hidden_layer_2_size_rpd[clust1_mask_rpd]
clust2_hidden_layer_2_rpd = hidden_layer_2_size_rpd[clust2_mask_rpd]
clust3_hidden_layer_2_rpd = hidden_layer_2_size_rpd[clust3_mask_rpd]

clust1_hidden_layer_3_rpd = hidden_layer_3_size_rpd[clust1_mask_rpd]
clust2_hidden_layer_3_rpd = hidden_layer_3_size_rpd[clust2_mask_rpd]
clust3_hidden_layer_3_rpd = hidden_layer_3_size_rpd[clust3_mask_rpd]

# ax21_2rpd.scatter(clust1_hidden_layer_2_rpd, clust1_adapter_accuracy_rpd, s=5, c='blue')
# ax21_2rpd.scatter(clust2_hidden_layer_2_rpd, clust2_adapter_accuracy_rpd, s=5, c='red')
ax1_3rpd.scatter(clust3_hidden_layer_2_rpd, clust3_adapter_accuracy_rpd, s=5, c='green')
ax1_3rpd.set_xlabel("Hidden Layer 2 Size")
ax1_3rpd.set_ylabel("Model Accuracy")
ax1_3rpd.set_title("Hidden Layer 2 Size vs Model Accuracy")

# ax22_2rpd.scatter(clust1_hidden_layer_3_rpd, clust1_adapter_accuracy_rpd, s=5, c='blue')
# ax22_2rpd.scatter(clust2_hidden_layer_3_rpd, clust2_adapter_accuracy_rpd, s=5, c='red')
ax2_3rpd.scatter(clust3_hidden_layer_3_rpd, clust3_adapter_accuracy_rpd, s=5, c='green')
ax2_3rpd.set_xlabel("Hidden Layer 3 Size")
ax2_3rpd.set_ylabel("Model Accuracy")
ax2_3rpd.set_title("Hidden Layer 3 Size vs Model Accuracy")

plt.tight_layout(pad=1.5, w_pad=1.5, h_pad=1.5)
fig3_rpd.set_dpi(500)


# The fourth figure
fig4_rpd, ((ax11_4rpd, ax12_4rpd), (ax21_4rpd, ax22_4rpd)) = plt.subplots(2, 2)

clust1_adapter_loss_rpd = adapter_loss_rpd[clust1_mask_rpd]
clust2_adapter_loss_rpd = adapter_loss_rpd[clust2_mask_rpd]
clust3_adapter_loss_rpd = adapter_loss_rpd[clust3_mask_rpd]

clust1_loss_function_rpd = loss_function_rpd[clust1_mask_rpd]
clust2_loss_function_rpd = loss_function_rpd[clust2_mask_rpd]
clust3_loss_function_rpd = loss_function_rpd[clust3_mask_rpd]

clust1_adapter_activation_rpd = adapter_activation_rpd[clust1_mask_rpd]
clust2_adapter_activation_rpd = adapter_activation_rpd[clust2_mask_rpd]
clust3_adapter_activation_rpd = adapter_activation_rpd[clust3_mask_rpd]

# ax11_3rpd.scatter(clust1_adapter_loss_rpd, clust1_adapter_accuracy_rpd, s=5, c='blue')
# ax11_3rpd.scatter(clust2_adapter_loss_rpd, clust2_adapter_accuracy_rpd, s=5, c='red')
ax11_4rpd.scatter(clust3_adapter_loss_rpd, clust3_adapter_accuracy_rpd, s=5, c='green')
ax11_4rpd.set_xlabel("Adapter Loss (Categorical Crossentropy)")
ax11_4rpd.set_ylabel("Model Accuracy")
ax11_4rpd.set_title("Adapter Loss vs Model Accuracy")

# ax12_3rpd.scatter(clust1_loss_function_rpd, clust1_adapter_accuracy_rpd, s=5, c='blue')
# ax12_3rpd.scatter(clust2_loss_function_rpd, clust2_adapter_accuracy_rpd, s=5, c='red')
ax12_4rpd.scatter(clust3_loss_function_rpd, clust3_adapter_accuracy_rpd, s=5, c='green')
ax12_4rpd.set_xlabel("Adapter Loss Function")
ax12_4rpd.set_ylabel("Model Accuracy")
ax12_4rpd.set_title("Loss Function vs Model Accuracy")

# ax21_3rpd.scatter(clust1_adapter_activation_rpd, clust1_adapter_accuracy_rpd, s=5, c='blue')
# ax21_3rpd.scatter(clust2_adapter_activation_rpd, clust2_adapter_accuracy_rpd, s=5, c='red')
ax21_4rpd.scatter(clust3_adapter_activation_rpd, clust3_adapter_accuracy_rpd, s=5, c='green')
ax21_4rpd.set_xlabel("Adapter Activation Funciton")
ax21_4rpd.set_ylabel("Model Accuracy")
ax21_4rpd.set_title("Activation Function vs Model Accuracy")

# ax22_3rpd.scatter(clust1_ensemble_accuracy_rpd, clust1_adapter_loss_rpd, s=5, c='blue')
# ax22_3rpd.scatter(clust2_ensemble_accuracy_rpd, clust2_adapter_loss_rpd, s=5, c='red')
ax22_4rpd.scatter(clust3_ensemble_accuracy_rpd, clust3_adapter_loss_rpd, s=5, c='green')
ax22_4rpd.set_xlabel("Ensemble Accuracy")
ax22_4rpd.set_ylabel("Model Loss")
ax22_4rpd.set_title("Ensemble Accuracy vs Model Loss")

plt.tight_layout(pad=1.5, w_pad=1.5, h_pad=1.5)
fig4_rpd.set_dpi(500)


# The fifth figure
fig5_rpd, (ax1_5rpd, ax2_5rpd) = plt.subplots(1, 2)

nonzero_ensemble_loss_mask_rpd = [False] * rows_rpd

clust1_adapter_loss_2_rpd_cp = adapter_loss_rpd[clust1_mask_rpd]
clust2_adapter_loss_2_rpd_cp = adapter_loss_rpd[clust2_mask_rpd]
clust3_adapter_loss_2_rpd_cp = adapter_loss_rpd[clust3_mask_rpd]

clust1_ensemble_loss_rpd_cp = ensemble_loss_rpd[clust1_mask_rpd]
clust2_ensemble_loss_rpd_cp = ensemble_loss_rpd[clust2_mask_rpd]
clust3_ensemble_loss_rpd_cp = ensemble_loss_rpd[clust3_mask_rpd]

clust1_adapter_accuracy_2_rpd_cp = adapter_accuracy_rpd[clust1_mask_rpd]
clust2_adapter_accuracy_2_rpd_cp = adapter_accuracy_rpd[clust2_mask_rpd]
clust3_adapter_accuracy_2_rpd_cp = adapter_accuracy_rpd[clust3_mask_rpd]

l = []
for i in range(len(clust1_ensemble_loss_rpd_cp)):
    if clust1_ensemble_loss_rpd_cp[i] == 0:
        l.append(i)
clust1_ensemble_loss_rpd = np.delete(clust1_ensemble_loss_rpd_cp, l)
clust1_adapter_loss_2_rpd = np.delete(clust1_adapter_loss_2_rpd_cp, l)
clust1_adapter_accuracy_2_rpd = np.delete(clust1_adapter_accuracy_2_rpd_cp, l)

l = []
for i in range(len(clust2_ensemble_loss_rpd_cp)):
    if clust2_ensemble_loss_rpd_cp[i] == 0:
        l.append(i)
clust2_ensemble_loss_rpd = np.delete(clust2_ensemble_loss_rpd_cp, l)
clust2_adapter_loss_2_rpd = np.delete(clust2_adapter_loss_2_rpd_cp, l)
clust2_adapter_accuracy_2_rpd = np.delete(clust2_adapter_accuracy_2_rpd_cp, l)

l = []
for i in range(len(clust3_ensemble_loss_rpd_cp)):
    if clust3_ensemble_loss_rpd_cp[i] == 0:
        l.append(i)
clust3_ensemble_loss_rpd = np.delete(clust3_ensemble_loss_rpd_cp, l)
clust3_adapter_loss_2_rpd = np.delete(clust3_adapter_loss_2_rpd_cp, l)
clust3_adapter_accuracy_2_rpd = np.delete(clust3_adapter_accuracy_2_rpd_cp, l)


# ax1_4rpd.scatter(clust1_ensemble_loss_rpd, clust1_adapter_loss_2_rpd, s=5, c='blue')
# ax1_4rpd.scatter(clust2_ensemble_loss_rpd, clust2_adapter_loss_2_rpd, s=5, c='red')
ax1_5rpd.scatter(clust3_ensemble_loss_rpd, clust3_adapter_loss_2_rpd, s=5, c='green')
ax1_5rpd.set_xlabel("Ensemble Loss (MSE)")
ax1_5rpd.set_ylabel("Adapter Loss (Categorical Crossentropy)")
ax1_5rpd.set_title("Ensemble Loss vs Model Loss")

# ax2_4rpd.scatter(clust1_ensemble_loss_rpd, clust1_adapter_accuracy_2_rpd, s=5, c='blue')
# ax2_4rpd.scatter(clust2_ensemble_loss_rpd, clust2_adapter_accuracy_2_rpd, s=5, c='red')
ax2_5rpd.scatter(clust3_ensemble_loss_rpd, clust3_adapter_accuracy_2_rpd, s=5, c='green')
ax2_5rpd.set_xlabel("Ensemble Loss (MSE)")
ax2_5rpd.set_ylabel("Model Accuracy")
ax2_5rpd.set_title("Ensemble Loss vs Model Accuracy")

plt.tight_layout(pad=1.5, w_pad=1.5, h_pad=1.5)
fig5_rpd.set_dpi(500)


# The sixth figure TODO runn test to hold data constant

# clust1_ensemble_size_rpd = ensemble_size_rpd[clust1_mask_rpd]
# clust2_ensemble_size_rpd = ensemble_size_rpd[clust2_mask_rpd]
clust3_ensemble_size_rpd = ensemble_size_rpd[clust3_mask_rpd]

fig6_rpd, (ax_6rpd) = plt.subplots(1, 1)

ax_6rpd.scatter(clust3_ensemble_size_rpd, clust3_adapter_accuracy_rpd, s=5)
ax_6rpd.plot((np.unique(clust3_ensemble_size_rpd)),
             np.poly1d(np.polyfit(clust3_ensemble_size_rpd, clust3_adapter_accuracy_rpd, 1))(np.unique(clust3_ensemble_size_rpd)))
ax_6rpd.set_xlabel("Ensemble Size")
ax_6rpd.set_ylabel("Model Accuracy")
ax_6rpd.set_title("Ensemble Size vs Model Accuracy")

plt.tight_layout(pad=1.5, w_pad=1.5, h_pad=1.5)
fig5_rpd.set_dpi(500)


# Graphing the results for the correlation test
ensemble_size_cd =       correlation_data[:, 0].astype(int)
adapter_activation_cd =  correlation_data[:, 1].astype(str)
hidden_layer_1_size_cd = correlation_data[:, 2].astype(int)
hidden_layer_2_size_cd = correlation_data[:, 3].astype(int)
hidden_layer_3_size_cd = correlation_data[:, 4].astype(int)
loss_function_cd =       correlation_data[:, 5].astype(str)
adapter_loss_cd =        correlation_data[:, 6].astype(float)
adapter_accuracy_cd =    correlation_data[:, 7].astype(float)
category_ratio_cd =      correlation_data[:, 8].astype(float)
ensemble_accuracy_cd =   correlation_data[:, 9].astype(float)
total_running_time_cd =  correlation_data[:, 10].astype(float)

rows_cd = correlation_data.shape[0]
clust1_mask = [False] * rows_cd
clust2_mask = [False] * rows_cd

for i in range(rows_cd):

    if adapter_accuracy_cd[i] < 0.625:
        clust1_mask[i] = True
    else:
        clust2_mask[i] = True

clust1_adapter_accuracy = adapter_accuracy_cd[clust1_mask]
clust2_adapter_accuracy = adapter_accuracy_cd[clust2_mask]
clust1_category_ratio = category_ratio_cd[clust1_mask]
clust2_category_ratio = category_ratio_cd[clust2_mask]

fig_cd, (ax11_cd, ax12_cd) = plt.subplots(1, 2)

ax11_cd.scatter(ensemble_accuracy_cd, adapter_accuracy_cd, s=5)
ax11_cd.set_xlabel("Ensemble Accuracy")
ax11_cd.set_ylabel("Model Accuracy")
ax11_cd.set_title("Ensemble Accuracy vs Model Accuracy")

ax12_cd.scatter(clust1_category_ratio, clust1_adapter_accuracy, s=5, c='red')
ax12_cd.scatter(clust2_category_ratio, clust2_adapter_accuracy, s=5, c='blue')
ax12_cd.set_xlabel("Category Ratio")
ax12_cd.set_ylabel("Model Accuracy")
ax12_cd.set_title("Category Ratio vs Model Accuracy")

plt.tight_layout(pad=1.5, w_pad=1.5, h_pad=1.5)


# SAVE THE FIGURES
fig1_rpd.savefig("../Figures/figure1.png", dpi=500)
fig2_rpd.savefig("../Figures/figure2.png", dpi=500)
fig3_rpd.savefig("../Figures/figure3.png", dpi=500)
fig4_rpd.savefig("../Figures/figure4.png", dpi=500)
fig5_rpd.savefig("../Figures/figure5.png", dpi=500)
fig6_rpd.savefig("../Figures/figure6.png", dpi=500)
fig_cd.savefig("../Figures/figure7.png", dpi=500)
