# results_parser
# Author:       Everett Harding Luke Buquicchio
# Date Created: 2/21/18
# Description:


import re
import matplotlib.pyplot as plt
import numpy as np
import glob


marker_size = 15
marker_color = (172./255., 43./255., 55./255.)


def represents_int(i):
    try:
        int(i)
        return True
    except ValueError:
        return False


def parse_dat_slurm(file):

    tests = []
    with open(file) as file:
        tree_data = False
        odd = True

        ensemble_sizes = 0

        temp_matrix = []
        loss_histories = []     # (training, testing) history of the adapter in this fold
        loss_history = []
        folds = []              # accumulated information in this fold
        precision_scores = []
        recall_scores = []
        f1_scores = []
        accuracy_scores = []
        confusion_matrices = []

        losses_accuracies = {}  # (losses for regressions on each continuous feature, model accuracy)
        featacc_accuracies = {}  # (accuracy for classification on each categorical feature, model accuracy)
        l_a_folds = {}
        f_a_folds = {}
        stddevs = {}            # standard deviations for each feature
        model_params = {}       # dictionary of parameters used in the model this fold

        for line in file:
            elems = re.split(':\t|\t| |\n|, |\[|: |\'|{|]|}', line)
            elems = [x for x in elems if x != '']

            if len(elems) == 0:
                continue

            if elems[0] == 'Randomized-Parameter-Test-Trees':
                tree_data = True

            elif elems[0] == 'Target' and elems[2] == 'Loss':
                ensemble_sizes += 1
                target = int(elems[1])
                loss = float(elems[3])
                stddev = float(elems[6])
                if target in l_a_folds.keys():
                    l_a_folds[target].append(loss)
                else:
                    l_a_folds[target] = [loss]
                stddevs[target] = stddev

            elif elems[0] == 'Target' and elems[2] == 'Accy':
                ensemble_sizes += 1
                target = int(elems[1])
                accuracy = float(elems[3])
                stddev = float(elems[6])
                if target in f_a_folds.keys():
                    f_a_folds[target].append(accuracy)
                else:
                    f_a_folds[target] = [accuracy]
                stddevs[target] = stddev

            elif represents_int(elems[0]) and represents_int(elems[1]) and len(elems) == 2 and odd:
                temp_matrix = [elems[0], elems[1]]
                odd = False

            elif represents_int(elems[0]) and represents_int(elems[1]) and len(elems) == 2 and not odd:
                confusion_matrices += [(temp_matrix[0], temp_matrix[1], elems[0], elems[1])]
                temp_matrix = []
                odd = True

            elif elems[0] in ['loss', 'loss:'] and len(elems) > 2:
                loss_history = []
                for x in elems[1:]:
                    loss_history.append(float(x))
                loss_histories.append(loss_history)

            elif elems[0] == 'acc' and len(elems) == 2:
                accuracy_scores += [float(elems[-1])]

            elif elems[0] == 'acc':
                acc_vals = []
                for x in elems[1:]:
                    acc_vals.append(x)

            elif len(elems) == 1 and tree_data:
                accuracy_scores.append(float(elems[0]))

            elif elems[0] == 'Precision':
                precision_scores += [float(elems[-1][:-1])] # hack because precision scores have an f appended to them in the file, whoops (it_was_a_typo)

            elif elems[0] == 'Recall':
                recall_scores += [float(elems[-1])]

            elif elems[0] == 'F1':
                f1 = float(elems[-1])
                f1_scores += [f1]

                # we've reached the end of a fold, store the fold and reset
                model_params['ensemble_size'] = ensemble_sizes

                # calculate average losses/accuracies

                for key in l_a_folds:
                    avg = 0
                    for l in l_a_folds[key]:
                        avg += float(l)
                    avg /= len(l_a_folds[key])

                    if key in losses_accuracies.keys():
                        losses_accuracies[key] += [(avg, f1)]
                    else:
                        losses_accuracies[key] = [(avg, f1)]

                for key in f_a_folds:
                    avg = 0
                    for l in f_a_folds[key]:
                        avg += float(l)
                    avg /= len(f_a_folds[key])

                    if key in featacc_accuracies.keys():
                        featacc_accuracies[key] += [(avg, f1)]
                    else:
                        featacc_accuracies[key] = [(avg, f1)]

                folds.append((l_a_folds, f_a_folds, stddevs, loss_history, model_params, accuracy_scores))

                ensemble_sizes = 0
                loss_history = []
                l_a_folds = {}
                f_a_folds = {}
                stddevs = {}
                model_params = {}

            elif elems[0] in ['in_shape', 'hidden1', 'hidden2', 'hidden3', 'batch_size', 'epochs']:
                label = elems[0]
                model_params[label] = elems[1]

            elif elems[0] == 'Total' and elems[1] == 'Time':
                tests.append((folds, precision_scores, recall_scores, f1_scores, accuracy_scores, confusion_matrices))
                folds = []
                precision_scores = []
                recall_scores = []
                f1_scores = []
                accuracy_scores = []
                confusion_matrices = []

    return tests


def make_adapter_graphs(files):
    experiments = []

    for file in files:
        experiments += parse_dat_slurm(file)

    ensemble_sizes = []
    adapter_precisions = []
    adapter_recalls = []
    adapter_f1scores = []
    adapter_accuracies = []
    full_accuracies = []
    fold_confusion_matrices = []
    adapter_accuracy_training_history = []

    need_new_ensemble_data = True
    pick_lc = True

    for folds, precision_scores, recall_scores, f1_scores, accuracy_scores, confusion_matrices in experiments:
        for la, fa, stddevs, loss_history, model_params, acc_vals in folds:

            if need_new_ensemble_data:
                ensemble_sizes.append(model_params["ensemble_size"])
                need_new_ensemble_data = False

            if pick_lc:
                adapter_accuracy_training_history.append(acc_vals)

        pick_lc = False
        need_new_ensemble_data = True


        full_accuracies += [accuracy_scores]

        adapter_precisions.append(np.average(precision_scores))
        adapter_recalls.append(np.average(recall_scores))
        adapter_f1scores.append(np.average(f1_scores))
        adapter_accuracies.append(np.average(accuracy_scores))
        fold_confusion_matrices.append(confusion_matrices)


    rows = len(adapter_accuracies)
    valid_data = [False] * rows
    for i in range(rows):
        if adapter_accuracies[i] > 0.55:
            valid_data[i] = True


    valid_ensemble_sizes = np.asarray(ensemble_sizes)[valid_data]
    valid_adapter_precisions = np.asarray(adapter_precisions)[valid_data]
    valid_adapter_recalls = np.asarray(adapter_recalls)[valid_data]
    valid_adapter_f1scores = np.asarray(adapter_f1scores)[valid_data]
    valid_adapter_accuracies = np.asarray(adapter_accuracies)[valid_data]
    adapter_accuracy_training_history = np.asarray(adapter_accuracy_training_history, dtype=float)


    es_v_ap, (es_v_ap_u, es_v_ap_s) = plt.subplots(1, 2, figsize=(10, 5))

    # es_v_ap_poly = np.polyfit(valid_ensemble_sizes, valid_adapter_precisions, 1)
    # es_v_ap_trend_line = np.poly1d(es_v_ap_poly)
    #
    # es_v_ap_u.plot(valid_ensemble_sizes, es_v_ap_trend_line(valid_ensemble_sizes), "r--")
    es_v_ap_u.scatter(valid_ensemble_sizes, valid_adapter_precisions, s=marker_size, c=marker_color)
    es_v_ap_u.set_ylim([0, 1])
    es_v_ap_u.set_xlabel("Ensemble Size")
    es_v_ap_u.set_ylabel("Adapter Precision")
    es_v_ap_u.set_title("Ensemble Size vs Adapter Precision")

    es_v_ap_s.scatter(valid_ensemble_sizes, valid_adapter_precisions, s=marker_size, c=marker_color)
    es_v_ap_s.set_xlabel("Ensemble Size")
    es_v_ap_s.set_ylabel("Adapter Precision")
    es_v_ap_s.set_title("Ensemble Size vs Adapter Precision Scaled")
    plt.tight_layout(pad=1.5, w_pad=3, h_pad=1.5)

    es_v_ar, (es_v_ar_u, es_v_ar_s) = plt.subplots(1, 2, figsize=(10, 5))

    es_v_ar_u.scatter(valid_ensemble_sizes, valid_adapter_recalls, s=marker_size, c=marker_color)
    es_v_ar_u.set_ylim([0, 1])
    es_v_ar_u.set_xlabel("Ensemble Size")
    es_v_ar_u.set_ylabel("Adapter Recall")
    es_v_ar_u.set_title("Ensemble Size vs Adapter Recall")

    es_v_ar_s.scatter(valid_ensemble_sizes, valid_adapter_recalls, s=marker_size, c=marker_color)
    es_v_ar_s.set_xlabel("Ensemble Size")
    es_v_ar_s.set_ylabel("Adapter Recall")
    es_v_ar_s.set_title("Ensemble Size vs Adapter Recall Scaled")
    plt.tight_layout(pad=1.5, w_pad=3, h_pad=1.5)

    es_v_af1, (es_v_af1_u, es_v_af1_s) = plt.subplots(1, 2, figsize=(10, 5))

    es_v_af1_u.scatter(valid_ensemble_sizes, valid_adapter_f1scores, s=marker_size, c=marker_color)
    es_v_af1_u.set_ylim([0, 1])
    es_v_af1_u.set_xlabel("Ensemble Size")
    es_v_af1_u.set_ylabel("Adapter F1 Score")
    es_v_af1_u.set_title("Ensemble Size vs Adapter F1 Score")

    es_v_af1_s.scatter(valid_ensemble_sizes, valid_adapter_f1scores, s=marker_size, c=marker_color)
    es_v_af1_s.set_xlabel("Ensemble Size")
    es_v_af1_s.set_ylabel("Adapter F1 Score")
    es_v_af1_s.set_title("Ensemble Size vs Adapter F1 Score Scaled")
    plt.tight_layout(pad=1.5, w_pad=3, h_pad=1.5)

    es_v_aa, (es_v_aa_u, es_v_aa_s) = plt.subplots(1, 2, figsize=(10, 5))

    es_v_aa_u.scatter(valid_ensemble_sizes, valid_adapter_accuracies, s=marker_size, c=marker_color)
    es_v_aa_u.set_ylim([0, 1])
    es_v_aa_u.set_xlabel("Ensemble Size")
    es_v_aa_u.set_ylabel("Adapter Accuracy")
    es_v_aa_u.set_title("Ensemble Size vs Adapter Accuracy")

    es_v_aa_s.scatter(valid_ensemble_sizes, valid_adapter_accuracies, s=marker_size, c=marker_color)
    es_v_aa_s.set_xlabel("Ensemble Size")
    es_v_aa_s.set_ylabel("Adapter Accuracy")
    es_v_aa_s.set_title("Ensemble Size vs Adapter Accuracy Scaled")
    plt.tight_layout(pad=1.5, w_pad=3, h_pad=1.5)

    learning_curves, (lc_us) = plt.subplots(1, 1, figsize=(5, 5))

    adapter_accuracy_training_history = np.average(adapter_accuracy_training_history, axis=0)

    lc_us.plot(adapter_accuracy_training_history)
    lc_us.set_xlabel("Training Epochs")
    lc_us.set_ylabel("Accuracy")
    lc_us.set_title("Average Adapter Training History")
    plt.tight_layout(pad=1.5, w_pad=3, h_pad=1.5)

    es_v_ap.savefig("../Figures/Adapter_Precision.png", dpi=500)
    es_v_ar.savefig("../Figures/Adapter_Recall.png", dpi=500)
    es_v_af1.savefig("../Figures/Adapter_F1.png", dpi=500)
    es_v_aa.savefig("../Figures/Adapter_Accuracy.png", dpi=500)
    learning_curves.savefig("../Figures/Adapter_LearningCurves.png", dpi=500)


def make_tree_graphs(files):
    experiments = []

    for file in files:
        experiments += parse_dat_slurm(file)

    ensemble_sizes = []
    forest_precisions = []
    forest_recalls = []
    forest_f1scores = []
    forest_accuracies = []
    fold_confusion_matrices = []
    forest_accuracy_training_history = []

    need_new_ensemble_data = True
    pick_lc = True

    for folds, precision_scores, recall_scores, f1_scores, accuracy_scores, confusion_matrices in experiments:
        for la, fa, stddevs, loss_history, model_params, acc_vals in folds:

            if need_new_ensemble_data:
                ensemble_sizes.append(model_params["ensemble_size"])
                need_new_ensemble_data = False

            if pick_lc:
                forest_accuracy_training_history.append(acc_vals)

        pick_lc = False
        need_new_ensemble_data = True

        forest_precisions.append(np.average(precision_scores))
        forest_recalls.append(np.average(recall_scores))
        forest_f1scores.append(np.average(f1_scores))
        forest_accuracies.append(np.average(accuracy_scores))
        fold_confusion_matrices.append(confusion_matrices)

    es_v_fp, (es_v_fp_u, es_v_fp_s) = plt.subplots(1, 2, figsize=(10, 5))

    es_v_fp_u.scatter(ensemble_sizes, forest_precisions, s=marker_size, c=marker_color)
    es_v_fp_u.set_ylim([0, 1])
    es_v_fp_u.set_xlabel("Ensemble Size")
    es_v_fp_u.set_ylabel("Forest Precision")
    es_v_fp_u.set_title("Ensemble Size vs Forest Precision")

    es_v_fp_s.scatter(ensemble_sizes, forest_precisions, s=marker_size, c=marker_color)
    es_v_fp_s.set_xlabel("Ensemble Size")
    es_v_fp_s.set_ylabel("Forest Precision")
    es_v_fp_s.set_title("Ensemble Size vs Forest Precision Scaled")
    plt.tight_layout(pad=1.5, w_pad=3, h_pad=1.5)

    es_v_fr, (es_v_fr_u, es_v_fr_s) = plt.subplots(1, 2, figsize=(10, 5))

    es_v_fr_u.scatter(ensemble_sizes, forest_recalls, s=marker_size, c=marker_color)
    es_v_fr_u.set_ylim([0, 1])
    es_v_fr_u.set_xlabel("Ensemble Size")
    es_v_fr_u.set_ylabel("Forest Recall")
    es_v_fr_u.set_title("Ensemble Size vs Forest Recall")

    es_v_fr_s.scatter(ensemble_sizes, forest_recalls, s=marker_size, c=marker_color)
    es_v_fr_s.set_xlabel("Ensemble Size")
    es_v_fr_s.set_ylabel("Forest Recall")
    es_v_fr_s.set_title("Ensemble Size vs Forest Recall Scaled")
    plt.tight_layout(pad=1.5, w_pad=3, h_pad=1.5)

    es_v_ff1, (es_v_ff1_u, es_v_ff1_s) = plt.subplots(1, 2, figsize=(10, 5))

    es_v_ff1_u.scatter(ensemble_sizes, forest_f1scores, s=marker_size, c=marker_color)
    es_v_ff1_u.set_ylim([0, 1])
    es_v_ff1_u.set_xlabel("Ensemble Size")
    es_v_ff1_u.set_ylabel("Forest F1 Score")
    es_v_ff1_u.set_title("Ensemble Size vs Forest F1 Score")

    es_v_ff1_s.scatter(ensemble_sizes, forest_f1scores, s=marker_size, c=marker_color)
    es_v_ff1_s.set_xlabel("Ensemble Size")
    es_v_ff1_s.set_ylabel("Forest F1 Score")
    es_v_ff1_s.set_title("Ensemble Size vs Forest F1 Score Scaled")
    plt.tight_layout(pad=1.5, w_pad=3, h_pad=1.5)

    es_v_fa, (es_v_fa_u, es_v_fa_s) = plt.subplots(1, 2, figsize=(10, 5))

    es_v_fa_u.scatter(ensemble_sizes, forest_accuracies, s=marker_size, c=marker_color)
    es_v_fa_u.set_ylim([0, 1])
    es_v_fa_u.set_xlabel("Ensemble Size")
    es_v_fa_u.set_ylabel("Forest Accuracy")
    es_v_fa_u.set_title("Ensemble Size vs Forest Accuracy")

    es_v_fa_s.scatter(ensemble_sizes, forest_accuracies, s=marker_size, c=marker_color)
    es_v_fa_s.set_xlabel("Ensemble Size")
    es_v_fa_s.set_ylabel("Forest Accuracy")
    es_v_fa_s.set_title("Ensemble Size vs Forest Accuracy Scaled")
    plt.tight_layout(pad=1.5, w_pad=3, h_pad=1.5)

    learning_curves, (lc_us) = plt.subplots(1, 1, figsize=(5, 5))

    forest_accuracy_training_history = np.average(forest_accuracy_training_history, axis=0)

    lc_us.plot(forest_accuracy_training_history)
    lc_us.set_xlabel("Training Epochs")
    lc_us.set_ylabel("Accuracy")
    lc_us.set_title("Average Forest Training History")
    plt.tight_layout(pad=1.5, w_pad=3, h_pad=1.5)

    es_v_fp.savefig("../Figures/Forest_Precision.png", dpi=500)
    es_v_fr.savefig("../Figures/Forest_Recall.png", dpi=500)
    es_v_ff1.savefig("../Figures/Forest_F1.png", dpi=500)
    es_v_fa.savefig("../Figures/Forest_Accuracy.png", dpi=500)
    learning_curves.savefig("../Figures/Forest_LearningCurves.png", dpi=500)


if __name__ == "__main__":
    adapter_files = glob.glob('../Results/Generation4/*.out')
    tree_files = glob.glob('../Results/TreeResults/*.out')

    make_tree_graphs(tree_files)
    make_adapter_graphs(adapter_files)
